'''
Feedforward neural network with some number of layers.
'''

from utils import *

NO_TARGET = 'x'

## RANDOM WEIGHT AND ACTIVATION GENERATION

def random_act(min_act, max_act):
    '''Generate a random activation between min_act and max_act.'''
    return min_act + random.random() * (max_act - min_act)

def random_weight(wt_range, neg=True):
    '''Generate a random weight with wt_range. If neg is True, value may be negative.'''
    val = random.random() * wt_range
    if neg:
        val -= wt_range / 2
    return val

class Network(object):

    LR = .05

    def __init__(self, name, layers, supervised=True):
        '''Initialize, creating input, hidden, and output Layers and weights.'''
        self.name = name
        # A list of layers in input - hidden - output order
        self.layers = layers
        # By default this is a supervised pattern associator
        self.supervised = supervised
        self.has_bias = True
        self.competitive = False
        self.settling = False
        self.init = False
        self.target = []
        self.initialize()

    def __str__(self):
        '''Print name for the Network.'''
        return self.name

    def initialize(self):
        '''Initialize the Layers, including weights to non-input Layers.'''
        for i in range(1, len(self.layers)):
            # Default connectivity: each layer feeds into the next
            self.layers[i].input_layer = self.layers[i-1]
            self.layers[i-1].output_layer = self.layers[i]
        for l in self.layers:
            l.initialize()

    def reinit(self):
        '''Reinitialize; only initialize the weights into layers other than input.'''
        for l in self.layers[1:]:
            l.reinit()

    def get_n_weights(self):
        '''Number of weights and biases in the network.'''
        return sum([l.get_n_weights() for l in self.layers])

    def assign_weights(self, weights):
        '''Assign a list of weights (as long as it lasts) to the connections in the network.'''
        index = 0
        for l in self.layers:
            if l.weights:
                for u in range(l.size):
                    for w in range(len(l.weights[u])):
                        l.weights[u][w] = weights[index]
                        index += 1

    def propagate_forward(self):
        '''Propagate activation forward through the network.'''
        for l in self.layers[1:]:
            l.update()

    def propagate_backward(self):
        '''Propagate error backward through the network.'''
        for l in reversed(self.layers[1:-1]):
            l.update_error()

    def update_weights(self, lr=None):
        '''Update the weights into each layer other than the first.'''
#        lr = lr or Network.LR
        for l in reversed(self.layers[1:]):
            l.learn(lr=lr)

    def step(self, pattern, train=True, show_act=False, lr=None):
        '''Run the network on one pattern, returning the error and [] (winner for CompetNetwork).'''
        error = 0.0
        if self.init:
            self.pre_stuff(self.target if self.supervised else [])
        else:
            self.init = True
        # Clamp the input Layer to the list within the input part of the pattern.
        self.layers[0].clamp(pattern[0] if self.supervised else pattern)
        # Update the other layers in sequence
        self.propagate_forward()        
        # Figure the error into each output unit, given a target sublist
        if self.supervised and pattern[1]:
            error += self.layers[-1].do_errors(pattern[1])
            self.target = pattern[1]
        if train:
            # If we're training, propagate error back
            self.propagate_backward()
            # Update weights
            self.update_weights(lr=lr)
        if show_act:
            self.show()
#        self.post_stuff(pattern[1] if self.supervised else [])
        return error, []

    def pre_stuff(self, target):
        pass

##    def post_stuff(self, target):
##        pass

    def show(self):
        '''Print input pattern and output activation.'''
        self.layers[0].show()
        self.layers[-1].show()

    def show_weights(self):
        '''Print the weights and biases in the newtork.'''
        for l in self.layers[1:]:
            l.show_weights()

class Layer(object):
    '''A list of units (or groups of units), each with an activation.'''

    # Parameter for the sigmoid activation function
    threshold = 0.0

    # Constants for activation functions
    LINEAR = 0
    SIGMOID = 1
    TANH = 2

    def __init__(self, name,
                 size=10,           # number of units
                 weight_range=.5,   # range of initial weights
                 has_bias=True,     # is there a bias into each unit?
                 neg_weights=True,  # can initial weights be negative?
                 groups=False,      # not implemented yet
                 labels=[],         # not implemented yet
                 gain=1.0,          # gain for the activation function (if non-linear)
                 lr=None,           # learning rate
                 linear=False,      # is the activation function linear?
                 bipolar=False,     # are activations negative and positive?
                 dimensions=None):  # width and height of a rectangle of units
                                    # (only relevant for competitive learning)
        '''Initialize variables, but not the list of activations or weights.'''
        self.dimensions = dimensions
        if dimensions:
            self.size = self.dimensions[0] * self.dimensions[1]
            self.set_coords()
        else:
            self.size = size
        # Layer feeding this layer
        self.input_layer = []
        # Layer fed from this layer
        self.output_layer = []
        self.weights = []
        # Whether to make a bias unit
        self.has_bias = has_bias
        # Whether to create initial negative weights
        self.neg_weights = neg_weights
        self.errors = []
        self.gain = gain
        self.lr = lr or .05
        # Activation function
        self.act_func = Layer.LINEAR if linear else (Layer.TANH if bipolar else Layer.SIGMOID)
        self.name = name
        self.min_activation = 0.0
        self.max_activation = 1.0
        self.weight_range = weight_range
        self.index = 0
        self.labels = labels

    def __str__(self):
        '''Print name for Layer.'''
        return self.name

    def set_coords(self):
        '''If units are arranged in a rectangle, set the coordinates for each unit.'''
        self.coords = {}
        unit = 0
        for y in range(self.dimensions[0]):
            for x in range(self.dimensions[1]):
                self.coords[unit] = x, y
                unit += 1

    def get_coords(self, unit):
        '''The rectangular coordinates of the unit.'''
        return self.coords[unit]

    def coords2index(self, coords):
        '''For units arranged in a rectangle, get the unit index from the coords of a unit.'''
        return coords[1] * self.dimensions[1] + coords[0]

    def get_neighbors(self, unit, rightdown=False):
        '''For units arranged in a rectangle, get the neighbor indices of one, only right and down if rightdown.'''
        coords = self.get_coords(unit)
        neighbors = []
        if coords[0] > 0 and not rightdown:
            # left neighbor
            neighbors.append(self.coords2index((coords[0] - 1, coords[1])))
        if coords[0] < self.dimensions[1] - 1:
            # right neighbor
            neighbors.append(self.coords2index((coords[0] + 1, coords[1])))
        if coords[1] > 0 and not rightdown:
            # upper neighbor
            neighbors.append(self.coords2index((coords[0], coords[1] - 1)))
        if coords[1] < self.dimensions[0] - 1:
            # lower neighbor
            neighbors.append(self.coords2index((coords[0], coords[1] + 1)))
        return neighbors

    def initialize(self):
        '''Create the activations and weights lists.'''
        self.activations = self.gen_random_acts()
        if self.input_layer:
            # This has an input_layer, so it has errors
            self.initialize_weights()
            self.errors = [0.0 for u in range(self.size)]

    def initialize_weights(self):
        '''Create the weights list of lists and a bias for each unit (if self.has_bias is True).

        Indices: [dest-index][input-unit-index]
        '''
        self.weights = [ [random_weight(self.weight_range, self.neg_weights) for i in range(self.input_layer.size)] \
                         # Bias weight
                         + ([random_weight(self.weight_range, self.neg_weights)] if self.has_bias else []) \
                         for u in range(self.size) ]

    def get_n_weights(self):
        '''Number of weights and biases into the layer.'''
        if self.weights:
            return sum([len(self.weights[u]) for u in range(self.size)])
        else:
            return 0

    def reinit(self):
        '''Reinitialize the Layer by assigning new weights and biases.'''
        for u in range(self.size):
            for j in range(len(self.weights[u])):
                self.weights[u][j] = random_weight(self.weight_range, self.neg_weights)

    def gen_random_acts(self):
        '''Generate random activations.'''
        return [random_act(self.min_activation, self.max_activation) for u in range(self.size)]

    def get_weights(self, dest_i, forward=True, limit=0):
        '''Get weights into (if forward) or out of unit dest_i.'''
        weights = self.weights[dest_i] if forward else self.get_reverse_weights(dest_i)
        if limit:
            # Only weights from limit inputs (or outputs if forward=False)
            weights = weights[:limit]
        return weights

    def get_reverse_weights(self, dest_i):
        '''Weights out of dest_i into output_layer.'''
        return [out[dest_i] for out in self.output_layer.weights] if self.output_layer else []

    def clamp(self, v):
        '''Clamp pattern vector v on this Layer.'''
##        if len(v) != self.size:
##            print 'Vector', v, 'is the wrong size'
##        else:
        for i in range(min([self.size, len(v)])):
            self.activations[i] = v[i]

    def do_errors(self, target):
        '''Figure the errors for each (output) unit, given the target pattern (list), returning RMS error.'''
        error = 0.0
        for i in range(self.size):
            e = 0.0 if (target[i] == NO_TARGET) else (target[i] - self.activations[i])
            self.errors[i] = e
            error += e * e
        return math.sqrt(error / self.size)

    def learn_incr(self, error, slope, src_act, lr):
        '''How much to increase a weight.'''
#        print 'Updating weight, LR', lr, 'slope', slope, 'error', error, 'act_func', self.act_func
        return error * slope * src_act * lr

    def learn(self, lr=None):
        '''Update the weights.'''
        lr = lr or self.lr
        for u in range(self.size):
            error = self.errors[u]
            # Slope depends on activation function
            if self.act_func == Layer.LINEAR:
                act_slope = 1.0
            elif self.act_func == Layer.SIGMOID:
                act_slope = sigmoid_slope(self.activations[u])
            else:
                act_slope = tanh_slope(self.activations[u])
            for i in range(self.input_layer.size):
                self.weights[u][i] += self.learn_incr(error, act_slope, self.input_layer.activations[i], lr)
            # Bias weight
            self.incr_bias(u, self.learn_incr(error, act_slope, 1.0, lr))

    def get_input(self, dest_i):
        '''Get input into unit dest_i, including [1.0] for bias.'''
        return dot_product(self.input_layer.activations + [1.0], self.weights[dest_i])

    def get_error_input(self, dest_i):
        '''Get the error input from the next layer into unit dest_i.'''
        return dot_product(self.output_layer.errors, self.get_reverse_weights(dest_i))

    def update_error(self):
        '''Update unit errors.'''
        for index in range(self.size):
            self.errors[index] = self.get_error_input(index)

    def update(self):
        '''Update unit activations.'''
        for index in range(self.size):
            inp = self.get_input(index)
            # Apply activation function
            if self.act_func == Layer.LINEAR:
                self.activations[index] = inp
            elif self.act_func == Layer.SIGMOID:
                self.activations[index] = sigmoid(inp, self.threshold, self.gain)
            else:
                self.activations[index] = math.tanh(self.gain * (-inp + self.threshold))

    def get_bias(self, unit):
        '''The bias into unit.'''
        return self.weights[u][self.input_layer.size]

    def set_bias(self, unit, bias):
        '''Set the bias into unit.'''
        self.weights[unit][self.input_layer.size] = bias

    def incr_bias(self, unit, incr):
        '''Increment the bias into unit.'''
        self.weights[unit][self.input_layer.size] += incr

    def show(self):
        '''Print activations.'''
        print self.name.ljust(12),
        for a in self.activations:
            print '%.3f' % a,
        print

    def show_weights(self):
        '''Print the weights and biases in the layer.'''
        print self.name
        for u in range(self.size):
            print str(u).ljust(5),
            for w in range(len(self.weights[u])):
                print '%.3f' % self.weights[u][w],
            print
