from Tkinter import *
import tkFont
import network

# Dimensions of the window for NN experiments
WIDTH = 600
HEIGHT = 500
MAX_CELL = 75
# Dimensions of individual windows for population
INDIV_WIDTH = 350
INDIV_HEIGHT = 300
# Number of times the Experiment is stepped when the 'Run' button is clicked.
N_STEPS = 10000
# Number of microseconds between steps during run.
STEP_DELAY = 0

def strength_color(strength, rgb):
    '''Convert a list of ints representing a color and a strength between 0 and 1 to a color.'''
    return "#%02x%02x%02x" % (int(strength * rgb[0]), int(strength * rgb[1]), int(strength * rgb[2]))

def get_node_color(activation):
    '''Get a color for a node with a particular activation: shades of green for +, red for -.'''
    if activation < 0.0:
        base = (255, 0, 0)
        activation = abs(activation)
    else:
        base = (0, 0, 255)
    if activation > 1.0:
        activation = 1.0
    return strength_color(activation, base)

def get_error_color(error):
    '''Get a color for an error value, shades of yellow.'''
    return strength_color(min([error, 1.0]), (255, 255, 0))

class NNTop(Toplevel):
    
    def __init__(self, width=INDIV_WIDTH, height=INDIV_HEIGHT,
                 critter = None, network = None, title = None, menus=True):
        Toplevel.__init__(self)
        self.critter = critter
        self.frame = NNFrame(self, critter = critter, network = network,
                             width = width, height = height,
                             buttons = False, menus = menus,
                             title = title)
        self.canvas = self.frame.canvas

    def destroy(self):
        '''When we destroy the network window, we need to tell its critter.'''
        self.critter.net_view = None
        Toplevel.destroy(self)

    def paint(self):
        '''Paint all nodes in the canvas.'''
        self.canvas.paint_nodes()

class NNFrame(Frame):
    '''A window for the Canvas showing the Network.'''

    def __init__(self, parent, experiment = None, network = None,
                 width = WIDTH, height = HEIGHT,
                 critter = None,
                 buttons = True, menus = True, title = None):
        '''Create the buttons and the Canvas.  Give it (and the Canvas) an Experiment or a Critter.'''
        Frame.__init__(self, parent, width=width, height=height)
        self.grid(row = 0)
        parent.title(title if title else "Biomorph")
        self.always_paint = False
        self.show_input_space = False
        self.canvas = NNCanvas(self, experiment, width = width, height = height,
                               network = network, critter = critter)
        if buttons:
            self.__button0 = Button(self, text = "Reinit")
            self.__button0.grid(row = 2, column = 0)
            self.__button0.bind("<Button-1>", self.canvas.reinit)
            self.__button1 = Button(self, text = "Step")
            self.__button1.grid(row = 2, column = 1)
            self.__button1.bind("<Button-1>", self.canvas.step)
            self.__button2 = Button(self, text = "Run")
            self.__button2.grid(row = 2, column = 2)
            self.__button2.bind("<Button-1>", self.canvas.run)
            self.__button3 = Button(self, text = "Next cond")
            self.__button3.grid(row = 2, column = 3)
            self.__button3.bind("<Button-1>", self.canvas.next_cond)

        if menus:
            self.viewmenu = Menubutton(self, text = "View")
            self.viewmenu.grid(row = 0, columnspan = 4)
            self.viewmenu.menu = Menu(self.viewmenu)
            self.viewmenu["menu"] = self.viewmenu.menu
            if self.canvas.is_nodes:
                self.viewmenu.menu.add_command(label = "Show hidden units in input space",
                                               command = self.__toggle_show_input_space)

    def __toggle_show_input_space(self):
        '''Display hidden nodes in input space (only when there are two dimensions).'''
        if self.show_input_space:
            self.viewmenu.menu.entryconfigure(1, label = "Show hidden units in input space")
            self.show_input_space = False
            self.canvas.unshow_input_space()
        else:
            self.viewmenu.menu.entryconfigure(1, label = "Show hidden units normally")
            self.show_input_space = True
            self.canvas.show_input_space()

class NNCanvas(Canvas):
    '''The Canvas where the Network is shown.'''

    node_radius = 1

    def __init__(self, parent, experiment = None, critter = None,
                 network = None, width = WIDTH, height = HEIGHT):
        Canvas.__init__(self, parent, width=width, height=height, bg="black")
        self.__width = width
        self.__height = height
        self.__parent = parent
        self.__experiment = experiment
        if experiment:
            self.__network = experiment.network
            self.__experiment.canvas = self
        else:
            self.__critter = critter
            self.__network = network
        self.__compet = self.__network.competitive
        self.__is_sup = self.__network.supervised
        self.__settling = self.__network.settling
        self.__bias = self.__network.has_bias
        self.grid(row = 1, columnspan = 4)
        self.__label_font = tkFont.Font(family='Helvetica', size = 9, weight = 'normal')
        self.__initialize()

    def __initialize(self):
        '''Create all of the objects for the nodes, weights, and targets, plus various lines.'''
        # Width and height of region where units are displayed
        width = 4 * self.__width / 5
        height = 7 * self.__height / 8
        y0 = self.__height / 16
        y1 = self.__height - y0
        # Lists and dict to store the objects
        self.__layers = self.__network.layers
        self.__dim_layers = len(filter(lambda l: l.dimensions != None, self.__layers))
        # Number of layers
        self.__n_layers = len(self.__layers)
        # Number of layer regions
        self.__n_regions = 2 * self.__dim_layers + (self.__n_layers - self.__dim_layers)
        # Height of layer region; depends on number of layers (+ target if supervised)
        layer_height = height / (self.__n_regions + (.5 if self.__is_sup else 0))
        max_cell = min(3 * layer_height / 4, MAX_CELL)
        layer_gap = layer_height / 8
        self.__units = []
        self.__labels = []
        self.__weights = []
        self.__targets = []
        self.__weight_dests = {}
        # Number of units
        self.__n_units = [l.size for l in self.__network.layers]

        # Error object: put in the upper right corner
        cell = self.__width / 30
        gap = cell / 2
        self.error_id = self.__add_graphic(self.__width - cell - gap, cell + gap, cell,
                                           outline='blue', fill='', rect=False)

        # Bias object: put in the lower left corner
        if self.__bias:
            cell = self.__width / 30
            gap = cell / 2
            self.bias_id = self.__add_graphic(cell + gap, self.__height - cell - gap, cell,
                                              outline='', fill='')

        for l, layer in enumerate(self.__layers):
            is_out = (l == self.__n_layers - 1)
            is_in = (l == 0)
            has_dims = (layer.dimensions != None)
            layer_width = layer.dimensions[0] if has_dims else layer.size
            x_gap = width / layer_width
            y_gap = (2 * layer_height) / layer.dimensions[1] if has_dims else 0
            if has_dims:
                x_gap = y_gap
            cell = min(x_gap / 4, max_cell / 2)
            x0 = (self.__width - (x_gap * layer_width)) / 2
            x = x0 + x_gap / 2
            y0 = (y1 - 2 * layer_height + cell + layer_gap) if has_dims else (y1 - cell - layer_gap)
            y = y0
            if not is_out:
                weights_y = y - 2 * cell - layer_gap
            elif self.__is_sup:
                target_y = y - 2 * cell - 2 * layer_gap
            units = []
            # List of weight icons OUT of this layer
            weights = []
            targets = []
            labels = layer.labels
            def __weights_hider(event, layer_index = l-1):
                return self.__unpaint_weights(event, layer_index)

            if has_dims and not is_in: #  and self.__layers[l - 1].size == 2:
                # Make an input space display of nodes and links
                self.create_is(x0, y0, layer)
            else:
                self.is_nodes = []

            for n in range(layer.size):
                unit_id = self.__add_graphic(x, y, cell)
                units.append(unit_id)
                if labels:
                    label_id = self.__add_label(labels[n], x, y)
                    self.__labels.append(label_id)
                if not is_out:
                    weight_id = self.__add_graphic(x, y, cell, outline='', fill='')
                    weights.append(weight_id)
                elif self.__is_sup:
                    target_id = self.__add_graphic(x, target_y, cell, '', '')
                    targets.append(target_id)
                    # Line between outputs and target
                    self.create_line(x0, y - cell - layer_gap, x0 + x_gap * layer_width, y - cell - layer_gap,
                                     fill='white')
                if not is_in:
                    # Handle displaying weights INTO this unit
                    def __weights_shower(event, unit_id = unit_id):
                        return self.__paint_weights(event, unit_id)
                    # Save this layer, to get the weights, and the index of the source layer
                    self.__weight_dests[unit_id] = (layer, l-1, n)
                    # Object that will get clicked to display the weights into the unit
                    button_unit_id = unit_id if is_out else weight_id
                    self.tag_bind(button_unit_id, "<Button-1>", __weights_shower)
                    self.tag_bind(button_unit_id, "<ButtonRelease-1>", __weights_hider)
                    if labels:
                        self.tag_bind(label_id, '<Button-1>', __weights_shower)
                        self.tag_bind(label_id, "<ButtonRelease-1>", __weights_hider)

                if has_dims and (n + 1) % layer.dimensions[0] == 0:
                    y += y_gap
                    x = x0 + x_gap / 2
                else:
                    x += x_gap

            self.__units.append(units)
            if not is_out:
                self.__weights.append(weights)
            elif self.__is_sup:
                self.__targets = targets
            y1 -= 2 * layer_height if has_dims else layer_height

    def create_is(self, x0, y0, layer):
        # Make an input space display of nodes and links
        self.ds_width = self.__width - 2 * x0
        self.x0 = x0
        self.y0 = y0
        self.is_nodes = []
        self.is_links = []
        self.weight_func = lambda x: layer.get_weights(x, limit=2)
        # Create the nodes
        for u in range(layer.size):
            weights = self.weight_func(u)  # layer.get_weights(u)
            x, y = weights[0] * self.ds_width + x0, weights[1] * self.ds_width + y0
            is_id = self.__add_is_node(x, y)
            self.is_nodes.append([is_id, (x, y)])
        # Create the links
        for u, node in enumerate(self.is_nodes):
            center_x, center_y = node[1]
            neighbors = layer.get_neighbors(u, True)
            for n in neighbors:
                other_x, other_y = self.is_nodes[n][1]
                link_id = self.__add_is_link(center_x, center_y, other_x, other_y)
                self.is_links.append([link_id, (u, n)])
        # Move the links below the nodes
        self.tag_lower('link', 'node')
        # Make the outside rectangle
        self.dim_square = self.create_rectangle(x0, y0,
                                                x0 + self.ds_width, y0 + self.ds_width,
                                                width = 1, outline = '', fill = '')

    def is_displayable(self, layer, last_layer):
        return layer.dimensions and last_layer #  and last_layer.size == 2
            
    def next_cond(self, event):
        '''Switch the Experiment to its next condition.'''
        self.__experiment.next_condition()

    ## A bunch of methods for displaying hidden layer in "input state"

    def __add_is_node(self, x, y):
        return self.create_oval(x - self.node_radius, y - self.node_radius,
                                x + self.node_radius, y + self.node_radius,
                                width = 0, fill = '', tags = ('node',))

    def __add_is_link(self, x1, y1, x2, y2):
        return self.create_line(x1, y1, x2, y2, fill = '', tags = ('link',))

    def unhide_is_nodes(self):
        for i in self.is_nodes:
            self.__unhide_is_node(i[0])

    def hide_is_nodes(self):
        for i in self.is_nodes:
            self.__hide_is_node(i[0])

    def unhide_is_links(self):
        for i in self.is_links:
            self.__unhide_is_link(i[0])

    def hide_is_links(self):
        for i in self.is_links:
            self.__hide_is_link(i[0])

    def __unhide_is_node(self, is_id):
        self.itemconfigure(is_id, fill = 'yellow')

    def __hide_is_node(self, is_id):
        self.itemconfigure(is_id, fill = '')

    def __unhide_is_link(self, link_id):
        self.itemconfigure(link_id, fill = 'blue')

    def __hide_is_link(self, link_id):
        self.itemconfigure(link_id, fill = '')

    def update_is(self):
        for u, i in enumerate(self.is_nodes):
            old_coords = i[1]
            weights = self.weight_func(u)
            new_coords = weights[0] * self.ds_width + self.x0, weights[1] * self.ds_width + self.y0
            self.coords(i[0], new_coords[0] - self.node_radius, new_coords[1] - self.node_radius,
                        new_coords[0] + self.node_radius, new_coords[1] + self.node_radius)
            i[1] = new_coords
        for l in self.is_links:
            node1, node2 = l[1]
            end_x1, end_y1 = self.is_nodes[node1][1]
            end_x2, end_y2 = self.is_nodes[node2][1]
            self.coords(l[0], end_x1, end_y1, end_x2, end_y2)

    def __add_graphic(self, x, y, w, outline = '#333', fill = 'green', rect = True):
        '''Add an object in the given position with outline and fill color, rectangle or circle.'''
        return \
               (getattr(self, 'create_rectangle') \
                if rect else getattr(self, 'create_oval'))(x - w, y - w, x + w, y + w,
                                                           width = 1, outline = outline, fill = fill)

    def __add_label(self, string, x, y, color = 'blue'):
        '''Add a text object for a unit.'''
        return self.create_text(x, y, text = string, fill = color, font = self.__label_font)

    def show_input_space(self):
        self.itemconfigure(self.dim_square, outline = 'yellow')
        self.hide_nodes()
        self.update_is()
        self.unhide_is_nodes()
        self.unhide_is_links()

    def unshow_input_space(self):
        self.itemconfigure(self.dim_square, outline = '')
        self.unhide_nodes()
        self.hide_is_nodes()
        self.hide_is_links()

    def paint_nodes(self):
        '''Paint all of the nodes.'''
        for l, layer in enumerate(self.__layers):
            if self.__parent.show_input_space and self.is_nodes and \
               l > 0 and self.is_displayable(layer, self.__layers[l - 1]):
                self.update_is()
            else:
                for u in range(layer.size):
                    self.__paint_node(self.__units[l][u], layer.activations[u])

    def __paint_node(self, nodeid, activation):
        '''Paint a node object.'''
        self.itemconfigure(nodeid, fill = get_node_color(activation))

    def hide_nodes(self):
        for l, layer in enumerate(self.__layers[1:]):
            if layer.dimensions:
                for u in range(layer.size):
                    self.__hide_node(self.__units[l + 1][u])

    def __hide_node(self, nodeid):
        self.itemconfigure(nodeid, fill = '', width = 0)

    def unhide_nodes(self):
        for l, layer in enumerate(self.__layers[1:]):
            if layer.dimensions:
                for u in range(layer.size):
                    self.__unhide_node(self.__units[l + 1][u], layer.activations[u])

    def __unhide_node(self, nodeid, activation):
        self.itemconfigure(nodeid, fill = get_node_color(activation), width = 1)

    def __paint_weights(self, event, unit_id):
        '''Show weights into unit represented by object with unit_id.'''
        dest_unit = self.__weight_dests[unit_id]
        layer = dest_unit[0]
        src_index = dest_unit[1]
        index = dest_unit[2]
        # Weights into dest unit, ending with bias
        weights = layer.get_weights(index)
        for w_i, w in enumerate(weights if not self.__bias else weights[:-1]):
            self.__paint_weight(src_index, w_i, w)
        # Bias
        if self.__bias:
            self.__paint_bias(weights[-1])
        print weights

    def __unpaint_weights(self, event, layer_index):
        '''Hide the weights for the source layer.'''
        for i in self.__weights[layer_index]:
            self.__unpaint_weight(i)
        if self.__bias:
            self.__unpaint_bias()

    def __paint_weight(self, layer_index, index, weight):
        '''Paint a weight object.'''
        self.itemconfigure(self.__weights[layer_index][index], fill = get_node_color(weight), outline = 'white')

    def __unpaint_weight(self, index):
        '''Unpaint (hide) the weight at index.'''
        self.itemconfigure(index, fill = '', outline = '')

    def __paint_bias(self, weight):
        '''Paint the bias into a unit.'''
        self.itemconfigure(self.bias_id, fill = get_node_color(weight), outline = 'white')

    def __unpaint_bias(self):
        self.itemconfigure(self.bias_id, fill = '', outline = '')

    def __paint_targets(self, target):
        '''Paint the targets (for a supervised Network).'''
        index = 0
        for u in range(len(target)):
            self.__paint_target(index, target[u])
            index += 1

    def __paint_target(self, index, target):
        '''Paint a target object.'''
        self.itemconfigure(self.__targets[index],
                           fill = get_node_color(0.0 if target == network.NO_TARGET else target),
                           outline = 'white')

    def __paint_error(self, error):
        self.itemconfigure(self.error_id, fill = get_error_color(error), outline = 'blue')

    def reinit(self, arg):
        '''Reinitialize the Network.'''
        self.__network.reinit()

    def step0(self, showError=True, paint=True, train=True, target=[]):
        '''Run the Experiment on one pattern.'''
        if self.__experiment:
            pat_err_win = self.__experiment.step(train=train, showError=showError)
            if self.__is_sup:
                target = pat_err_win[0][1]
            error = pat_err_win[1]
            if paint:
                self.__paint(target, error)
            return error, pat_err_win[2]
        elif paint:
            self.__paint(target)

    def __paint(self, target = [], error = None):
        self.paint_nodes()
        if error != None:
            self.__paint_error(error)
        if target:
            self.__paint_targets(target)

    def run(self, arg):
        '''Run the Experiment N_STEPS times and print out statistics.'''
        trial0 = self.__experiment.trials
        error = 0.0
        compet = False
        paint = self.__parent.always_paint
        if not self.__network.supervised:
            winners = set([])
            compet = True
        for i in range(N_STEPS):
            if paint:
                self.after(STEP_DELAY)
            err_win = self.step0(showError = False, paint = paint)
            error += err_win[0]
            if compet:
                winners = winners | err_win[1]
            if paint:
                self.update_idletasks()
        error /= N_STEPS
        print self.__experiment.trials, 'trials'
        print 'error:', error
        if self.__parent.show_input_space and self.is_nodes:
            self.update_is()
        self.__paint_error(error)
        if compet:
            print '# winners:', len(winners)

    def step(self, arg):
        '''Run the Experiment or Network n one pattern and show error.'''
        self.step0(showError = True)
