'''
Pattern and sequence generators for perception and action.
'''

from utils import *

# Pattern elements
DONT_CARE = 'x'

# Noise to add to patterns
NOISE = .05
DEFAULT_ACT = 0.0

def gen_random_value(continuous = True, bipolar = False):
    '''A random value for a pattern.'''
    val = ran(continuous)
    if bipolar:
        val = 2 * val - 1
    return val

def string2value(string, continuous = True, target = False, bipolar = False, squeeze = True):
    '''Convert a string to a pattern value.'''
    if string == DONT_CARE:
        return DONT_CARE
    else:
        val = float(string) if continuous else int(string)
        if squeeze:
            if val == 1:
                val = .9
            elif val == 0:
                val = .1
        return val

def string2number(string, continuous = True, bipolar = False):
    if string == DONT_CARE:
        return gen_random_value(continuous, bipolar)
    else:
        print "DON'T KNOW HOW TO CONVERT", string

class PatGen(object):
    '''A generator for patterns (lists).'''

    def __init__(self,
                 n,                                   # Length of input patterns
                 function = None,                     # Explicit function to call on each step
                 target = 0,                          # Length of target; if 0, no target
                 noise = 0.0, targ_noise = 0.0,       # Noise rates for input and target
                 ran_prob = 0.0,                      # Probability of a random pattern
                 squeeze = True,                      # Whether to keep targets away from asymptotes
                 continuous = True, bipolar = False,  # Kinds of input pattern values
                 default = DEFAULT_ACT,               # Default pattern value
                 filename = None,                     # File to read in patterns from
                 patterns = [],                       # Explicit list of patterns
                 sequential = False,                  # Whether patterns must appear in order
                 check = True):                       # Whether to check for errors as patterns are created
        '''Set pattern parameters and create patterns from patterns list or file.'''
        self.n = n
        self.function = function
        self.random = random
        self.noise = noise
        self.targ_noise = targ_noise
        self.continuous = continuous
        self.bipolar = bipolar
        self.ran_prob = ran_prob
        self.squeeze = squeeze
        self.min = -1 if bipolar else 0
        self.check = check
        self.sequential = sequential
        # Length of target pattern
        self.target = target
        self.gen_value = lambda: self.gen()
        if filename:
            self.read_patterns(filename)
        else:
            self.patterns = patterns
        if sequential:
            self.init_iterator()
        if check:
            self.check_patterns()

    def check_patterns(self):
        '''Check to see whether patterns are the right length.'''
        for p in self.patterns:
            inp = p[0] if self.target else p
            if len(inp) > self.n:
                print 'WARNING: input pattern', inp, 'is too long!'
            elif len(inp) > self.n:
                print 'WARNING: input pattern', inp, 'is too short!'
            if self.target:
                target = p[1]
                if len(target) > self.target:
                    print 'WARNING: target pattern', target, 'is too long!'
                elif len(target) > self.target:
                    print 'WARNING: target pattern', target, 'is too short!'

    def init_iterator(self):
        self.iterator = iter(self.patterns)

    def string2pattern(self, string_list, target):
        return [string2value(s, continuous = self.continuous, target = target, bipolar = self.bipolar, squeeze = self.squeeze) \
                for s in string_list]

    def read_patterns(self, filename):
        '''Read the patterns in from a file, ignoring blank lines.'''
        try:
            pat_file = open(filename)
            self.patterns = []
            inp = True
            for line in pat_file:
                pattern = self.string2pattern(line.split(), (not inp and self.target))
                if not self.target:
                    self.patterns.append(pattern)
                elif inp:
                    pat = [pattern]
                    inp = False
                else:
                    pat.append(pattern)
                    self.patterns.append(pat)
                    inp = True

            pat_file.close()
        except IOError:
            print 'No file named', filename

    def gen_random(self):
        '''A random pattern.'''
        return [gen_random_value(self.continuous, self.bipolar) for x in range(self.n)]

    def next(self, repeat = True):
        '''For sequential patterns, get the next one, resetting the iterator if at end.'''
        try:
            pattern = self.iterator.next()
            return pattern
        except StopIteration:
#            print 'End of sequence'
            self.init_iterator()
            return self.iterator.next() if repeat else False

    def gen(self, repeat = True):
        '''Generate a pattern.'''
        # Call the function
        if self.function:
            return self.function()
        # Generate a random pattern
        elif (self.ran_prob > 0.0 and random.random() < self.ran_prob) or not self.patterns:
            return self.gen_random()
        # Select from one of the predefined patterns
        else:
            pattern = self.next(repeat) if self.sequential else random.choice(self.patterns)
            return self.realize(pattern)

    def realize(self, pattern):
        '''Return copy of pattern, noisified and with DONT_CAREs filled in (unless target).'''
        inp = self.pattern_input(pattern)
        if self.noise > 0:
            self.noisify(inp)
            self.convert(inp)
        if self.target:
            targ = self.pattern_target(pattern)
            if self.targ_noise > 0:
                self.noisify(targ)
            return inp, targ
        return inp

    def convert(self, ls):
        '''Convert all non-numeric characters to numeric ones.'''
        for i, x in enumerate(ls):
            if isinstance(x, str):
                ls[i] = string2number(x, continuous = self.continuous, bipolar = self.bipolar)

    def pattern_input(self, pattern):
        '''A copy of the input part of the pattern.'''
        return pattern[0][:] if self.target else pattern[:]

    def pattern_target(self, pattern):
        '''A copy of the target part of the pattern; does not check whether this one.'''
        return pattern[1][:]

    def noisify(self, ls):
        '''Add noise to elements of ls.'''
        for i in range(len(ls)):
            if isinstance(i, str):
                # Don't do anything if there's a string here
                pass
            elif self.continuous:
                change = random.random() * self.noise
                if random.random() < .5:
                    change = -change
                    ls[i] = max(self.min, min(1.0, ls[i] + change))
                elif random.random() < self.noise:
                    if self.bipolar:
                        vals = [-1, 0, 1]
                        vals.remove(ls[i])
                        ls[i] = random.choice(vals)
                    else:
                        ls[i] = 0 if ls[i] else 1
            
    def __call__(self):
        '''What to do when the object is called.'''
        return self.gen(True)

class SeqGen(object):
    '''Generator for sequences of patterns.'''

    def __init__(self, pat_length, target = 0,   # No target if length = 0
                 noise = NOISE, targ_noise = NOISE,
                 ran_prob = .1, continuous = True, bipolar = False, default = DEFAULT_ACT,
                 # Sequences from file or lists passed directly to constructor
                 filename = None, sequences = [],
                 check = True):
        self.pat_length = pat_length
        self.target = target
        self.random = random
        self.noise = noise
        self.targ_noise = targ_noise
        self.continuous = continuous
        self.bipolar = bipolar
        self.ran_prob = ran_prob
        self.min = -1 if bipolar else 0
        self.check = check
        self.gen_value = lambda: self.gen()
        if filename:
            self.read_sequences(filename)
        else:
            self.sequences = [self.make_pat_gen(s) for s in sequences]
        self.init_iterator()
#        if check:
#            for s in self.sequences:
#                s.check_patterns()

    def make_pat_gen(self, patterns):
        '''Create a pattern generator for one sequence, given a list of patterns.'''
        return PatGen(n = self.pat_length, target = self.target,
                      noise = self.noise, ran_prob = self.ran_prob,
                      continuous = self.continuous, bipolar = self.bipolar,
                      patterns = patterns, sequential = True,
                      check = self.check)

    def init_iterator(self):
        '''Pick a sequence and initialize its iterator.'''
        # Pick a random sequence
        self.sequence = random.choice(self.sequences)
        self.sequence.init_iterator()

    def gen(self):
        '''Generate another pattern.'''
        next = self.sequence.next(repeat = False)
        if not next:
            # End of current sequence; go to a new one
            self.init_iterator()
            return self.sequence.next(repeat = False)
        else:
            return next

    def read_sequences(self, filename):
        '''Read in sequences from file, treating blank lines as separators.'''
        try:
            self.sequences = []
            seq_file = open(filename)
            pats = []
            inp = True
            for line in seq_file:
                split = line.split()
                if not split:
                    # Line is empty, end current seq
                    if not inp and self.target:
                        print 'ODD NUMBER OF LINES IN SEQUENCE, MISSING TARGET'
                        pats = pats[:-1]
                    self.sequences.append(self.make_pat_gen(pats))
                    pats = []
                else:
                    pattern = [string2value(s, continuous = self.continuous, bipolar = self.bipolar,
                                           target = (not inp and self.target)) \
                               for s in split]
                    if not self.target:
                        pats.append(pattern)
                    elif inp:
                        pat = [pattern]
                        inp = False
                    else:
                        pat.append(pattern)
                        pats.append(pat)
                        inp = True
            # End the last sequence
            if pats:
                self.sequences.append(self.make_pat_gen(pats))

            seq_file.close()

        except IOError:
            print 'No file named', filename

    def __call__(self):
        '''What to do when the object is called.'''
        return self.gen()

# Some examples of pattern generators

# XOR
xor_pg = PatGen(2, target=1, continuous=False,
                patterns = [[[0, 0], [0]], [[0, 1], [1]], [[1, 0], [1]], [[1, 1], [0]]])

# OR
or_pg = PatGen(2, target=1, continuous=False,
               patterns = [[[0, 0], [0]], [[0, 1], [1]], [[1, 0], [1]], [[1, 1], [1]]])

# 2 binary categories, length 4
cl_4_2_pg = PatGen(4, patterns = [[1, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 1], [0, 0, 1, 0]])

# Auto-associative local patterns of length 4
aa_4_pg = PatGen(4, patterns = [ [[1, 0, 0, 0], [1, 0, 0, 0]],
                                 [[0, 1, 0, 0], [0, 1, 0, 0]],
                                 [[0, 0, 1, 0], [0, 0, 1, 0]],
                                 [[0, 0, 0, 1], [0, 0, 0, 1]] ])

# Auto-associative local patterns of length 6
aa_6_pg = PatGen(6, patterns = [ [[1, 0, 0, 0, 0, 0], [1, 0, 0, 0, 0, 0]],
                                 [[0, 1, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0]],
                                 [[0, 0, 1, 0, 0, 0], [0, 0, 1, 0, 0, 0]],
                                 [[0, 0, 0, 1, 0, 0], [0, 0, 0, 1, 0, 0]],
                                 [[0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 1, 0]],
                                 [[0, 0, 0, 0, 0, 1], [0, 0, 0, 0, 0, 1]] ])

def gen_gauss():
    '''A number between 0 and 1 with a strong tendency to be near .5.'''
    return max([min([random.gauss(.5, .2), 1.0]), 0.0])

def gen_intensities(points, gauss = False):
    '''Distances of a random point in 2-space from each of points.'''
    if gauss:
        position = gen_gauss(), gen_gauss()
    else:
        position = random.random(), random.random()
    return [intensity(position, point, minimum = .05) for point in points]

# 2 random inputs
unit_square_pg = PatGen(2)

# 2 random inputs clustering around .5
unit_square_gauss_pg = PatGen(2, lambda: [gen_gauss(), gen_gauss()])

# "Intensities" of random points in 2-space from two "lights"
lights37_pg = PatGen(2, function = lambda: gen_intensities(((.3, .3), (.7, .7))))

lights01_pg = PatGen(2, function = lambda: gen_intensities(((0, 0), (1, 1))))

# "Intensities" of random points clustered around .5, .5 from two "lights"
lights37_gauss_pg = PatGen(2, function = lambda: gen_intensities(((.3, .3), (.7, .7)), True))

lights01_gauss_pg = PatGen(2, function = lambda: gen_intensities(((0, 0), (1, 1)), True))

# Sequence generators

aaabbb_in_sg = SeqGen(2, target = 2,
                      sequences = [ [ [ [1, 0], [1, 0] ],
                                      [ [1, 0], [1, 0] ],
                                      [ [1, 0], [0, 1] ],
                                      [ [0, 1], [0, 1] ],
                                      [ [0, 1], [0, 1] ],
                                      [ [0, 1], [1, 0] ]
                                      ]
                                    ])

aaabbb_out_sg = SeqGen(1, target = 2,
                       sequences = [ [ [ [1], [1, 0] ],
                                       [ [1], [1, 0] ],
                                       [ [1], [1, 0] ],
                                       [ [1], [0, 1] ],
                                       [ [1], [0, 1] ],
                                       [ [1], [0, 1] ]
                                       ]
                                     ])
